open X86

type var = Ast.var

(* simple let language intermediate representation -------------------------- *)
module IR = struct
  (* Unique identifiers for temporaries. *)
  type uid = int

  (* Labels *)
  type lbl = string

  (* A context maintains information for translating to/from the IR.
     In particular, used to allocate UIDs when translating to the IR
     and to manage stack layout when generating X86. *)
  type ctx =
    { mutable max_uid : int; (* number of allocated temporaries *)
      mutable max_label : int; (* number of allocated labels *)
      vars : var list (* list of (non-temporary, non-parameter) variables *)
    }

  (* Number of allocated temporaries *)
  let nb_uid c = c.max_uid

  (* Number of (non-temporary, non-parameter) variables *)
  let nb_vars c = List.length c.vars

  (* Allocate a new context *)
  let mk_context : var list -> ctx =
    fun vars -> { max_uid = 0; max_label = 0; vars = vars }
  
  (* Find the numerical position of a variable in a context *)
  let var_position ctx var =
    let rec go n xs =
      match xs with
      | (x::xs) ->
        if x = var then n else go (n + 1) xs
      | [] -> failwith (Printf.sprintf "Variable %s is not known to contex" var)
    in
    go 0 ctx.vars

  (* "gensym" -- generate a new unique identifier *)
  let mk_uid (c:ctx) : uid = 
    let uid = c.max_uid in
    c.max_uid <- c.max_uid + 1;
    uid

  let mk_label (c:ctx) (name:string) : lbl =
    let id = c.max_label in
    c.max_label <- c.max_label + 1;
    Printf.sprintf "%s.%d" name id

  (* operands *)
  type opn = 
    | Id of uid
    | Const of int64

  (* binary operations *)
  type bop =
    | Add
    | Mul
  
  (* instructions *)
  (* note that there is no nesting of operations! *)
  type instr =
    | Let of uid * bop * opn * opn
    | Load of uid * var
    | Store of var * opn

  type cond =
    | EqZ
    | LeZ
    | LtZ
  
  type terminator =
    | Ret of opn
    | Br of lbl                       (* unconditional branch *)
    | Cbr of cond * opn * lbl * lbl   (* conditional branch *)

  type elt =
    | L of lbl        (* Block labels *)
    | I of instr      (* Instructions *)
    | T of terminator (* Block terminators *)

  


  (* Basic blocks *)
  type block = { instrs: instr list; terminator: terminator }

  (* Control Flow Graph: a pair of an entry block and a set labeled blocks *)
  type cfg = block * (lbl * block) list


  (* Function declaration (currently unused) *)
  type fdecl = { name : string
               ; param : var list
               ; cfg : cfg }

  type program = fdecl list
  
  (* pretty printing *)
  let pp_uid u = Printf.sprintf "tmp%d" u
  let pp_var x = Printf.sprintf "var%s" x
  
  let pp_opn = function 
    | Id u   -> pp_uid u
    | Const c -> (Int64.to_string c)^"L"

  let pp_bop = function
    | Add -> "add"
    | Mul -> "mul"
  
  let pp_instr = function
    | Let (u, bop, op1, op2) ->
      Printf.sprintf "%s := %s %s %s"
        (pp_uid u) (pp_bop bop) (pp_opn op1) (pp_opn op2)
    | Load (u, x) ->
      Printf.sprintf "%s := load %s"
        (pp_uid u) (pp_var x)
    | Store (x, op) ->
      Printf.sprintf "store %s := %s"
        (pp_var x) (pp_opn op)
  
  let pp_program instrs =
    (String.concat "\n" (List.map pp_instr instrs))

  module MLMeaning = struct
    let add = Int64.add
    let mul = Int64.mul
    let load x = !x
    let store o x = x := o
  end
end

module Opt = struct
  open IR

  (* Redundant load elimination on a block *)
  let rle_block (b : block) =
    let subst_opn subst opn =
      match opn with
      | Id uid -> begin match List.assoc_opt uid subst with
          | Some uid' -> Id uid'
          | None -> Id uid
        end
      | Const k -> Const k
    in
    let (instrs, loaded, subst) =
      List.fold_left (fun (instrs, loaded, subst) instr ->
          match instr with
          | Load (uid, var) ->
            if List.mem_assoc var loaded then
              (instr::instrs,
               loaded,
               (uid, List.assoc var loaded)::subst)
            else
              (instr::instrs,
               (var, uid)::loaded,
               subst)
          | Store (var, _) ->
            (instr::instrs,
             List.remove_assoc var loaded,
             subst)
          | Let (uid, op, opn1, opn2) ->
            ((Let (uid, op, subst_opn subst opn1, subst_opn subst opn2))::instrs,
             loaded,
             subst))
        ([], [], [])
        b.instrs
    in
    { instrs = List.rev instrs;
      terminator = match b.terminator with
        | Ret opn -> Ret (subst_opn subst opn)
        | Br lbl -> Br lbl
        | Cbr (cond, opn, lbl1, lbl2) ->
          Cbr (cond, subst_opn subst opn, lbl1, lbl2) }

  let rle (entry, blocks) =
    (rle_block entry, List.map (fun (lbl, block) -> (lbl, rle_block block)) blocks)
end


(* When generating intermediate code, it's more efficient to emit code
   in *reverse* order, so that we can add to a code stream by
   prepending rather than appending.

   Q: If we generate code by repeatedly appending new code blocks to a
   stream, what is the time complexity of code generation?  What is
   the time complexity if we do it in reverse?

   The following functions reverse the order of append and cons to
   make the reversal look more natural. *)
let ( >@ ) x y = y @ x
let ( >:: ) x y = y :: x

(* Given a context c and expression e represented as an Ast, generate
   a pair (instrs, opn) such that after executing the sequence instr
   (in reverse order!), opn evaluates to the same value as e *)
let rec translate_exp (c : IR.ctx) (e : Ast.exp) : (IR.instr list * IR.opn) =
  let translate_bop bop e1 e2 =
    let (instrs1, opn1) = translate_exp c e1 in
    let (instrs2, opn2) = translate_exp c e2 in
    let id = IR.mk_uid c in
    IR.(instrs1 >@ instrs2 >@ [Let (id, bop, opn1, opn2)], Id id)
  in
  match e with
  | Var v ->
    let uid = IR.mk_uid c in
    IR.([Load (uid, v)], Id uid)
  | Const k -> IR.([], Const k)
  | Add (e1, e2) -> translate_bop IR.Add e1 e2
  | Mul (e1, e2) -> translate_bop IR.Mul e1 e2
  | Neg e -> translate_bop IR.Mul (Ast.Const (-1L)) e

let lift : (IR.instr list) -> IR.elt list = List.map (fun i -> IR.I i) 

let rec translate_cmd (ctx : IR.ctx) (cmd : Ast.cmd) : (IR.elt list) =
  let open Ast in
  match cmd with
  | Skip -> []
  | Assn (var, exp) ->
    let (instrs, opn) = translate_exp ctx exp in
    IR.((lift instrs) >@ [I (Store (var, opn))])
  | Seq (cmd1, cmd2) ->
    (translate_cmd ctx cmd1) >@ (translate_cmd ctx cmd2)
  | Return (exp) ->
    let (instrs, opn) = translate_exp ctx exp in
    IR.((lift instrs)
        >:: T (Ret opn))
  | WhileLZ (cond, body) ->
    let (instrs, opn) = translate_exp ctx cond in
    let loop_label = IR.mk_label ctx "loop" in
    let body_label = IR.mk_label ctx "body" in
    let exit_label = IR.mk_label ctx "exit" in
    IR.([T (Br loop_label)] (* Terminate prev block *)
        >:: L loop_label    (* Start of loop: evaluate guard & branch *)
        >@ lift instrs
        >:: T (Cbr (LtZ, opn, body_label, exit_label))

        >:: L body_label    (* loop body *)
        >@ translate_cmd ctx body
        >:: T (Br loop_label)
        >:: L exit_label)
  | IfLZ (cond, bthen, belse) ->
    let (instrs, opn) = translate_exp ctx cond in
    let then_label = IR.mk_label ctx "lz" in
    let else_label = IR.mk_label ctx "nlz" in
    let merge_label = IR.mk_label ctx "merge" in
    IR.((lift instrs)
        >:: T (Cbr (LtZ, opn, then_label, else_label))

        >:: L then_label
        >@ (translate_cmd ctx bthen)
        >:: T (Br merge_label)

        >:: L else_label
        >@ (translate_cmd ctx belse)
        >:: T (Br merge_label)
          
        >::L merge_label)

(* Convert an instruction stream into a control flow graph.  Assumes
   that the instructions are in 'reverse' order of execution.  *)
let build_cfg (code:(IR.elt list)) : IR.cfg =
  let open IR in
  let blocks_of_stream (code:elt list) =
    let (instrs, term_opt, blks) =
      List.fold_left
	(fun (instrs, term_opt, blks) e ->
           begin match e with
             | L l ->
               begin match term_opt with
                 | None ->
                   if (List.length instrs) = 0 then
                     ([], None, blks)
                   else failwith @@
                     Printf.sprintf "build_cfg: block labeled %s has\
                                     no terminator" l

                 | Some terminator ->
                   ([], None, (l, {instrs; terminator})::blks)
               end
             | T t  -> ([], Some t, blks)
             | I i  -> (i::instrs, term_opt, blks)
           end)
        ([], None, []) code
    in
    begin match term_opt with
      | None -> failwith "build_cfg: entry block has no terminator" 
      | Some terminator ->
        ({instrs; terminator}, blks)
    end
  in
  blocks_of_stream code 

(* Translate uid to indirect address *)
let codegen_uid uid =
  Ind3 (Lit (Int64.mul (-8L) (Int64.of_int (1 + uid))), Rbp)

(* Translate IR operand to X86 operand *)
let codegen_operand (operand : IR.opn) : X86.operand =
  let open Asm in
  let open IR in
  match operand with
  | Id uid -> codegen_uid uid
  | Const k -> Imm (Lit k)

(* Translate variable to X86 operand *)
let codegen_variable (c : IR.ctx) (v : var) : X86.operand =
  let open Asm in
  match v with
  | "x1" -> ~%Rdi
  | "x2" -> ~%Rsi
  | "x3" -> ~%Rdx
  | "x4" -> ~%Rcx
  | "x5" -> ~%R08
  | "x6" -> ~%R09
  | "x7" -> Ind3 (Lit 16L, Rbp)
  | "x8" -> Ind3 (Lit 24L, Rbp)
  | _ ->
    let pos = IR.var_position c v in
    let disp =
      Int64.mul (-8L) (Int64.of_int (1 + pos + (IR.nb_uid c)))
    in
    Ind3 (Lit disp, Rbp)

let codegen_instr (c : IR.ctx) (ins : IR.instr) : X86.ins list =
  match ins with
  | (IR.Let (uid, op, x, y)) ->
    let opcode = match op with
      | IR.Add -> Addq
      | IR.Mul -> Imulq
    in
    Asm.[Movq, [codegen_operand x; ~%Rax]
        ;Movq, [codegen_operand y; ~%R10]
        ;opcode, [~%R10; ~%Rax]
        ;Movq, [~%Rax; codegen_uid uid]]
  | IR.Load (uid, var) ->
    Asm.[Movq, [codegen_variable c var; ~%Rax]
        ;Movq, [~%Rax; codegen_uid uid]]
  | IR.Store (var, opn) ->
    Asm.[Movq, [codegen_operand opn; ~%Rax]
        ;Movq, [~%Rax; codegen_variable c var]]

let codegen_block (c : IR.ctx) block =
  let open IR in
  let terminator = match block.terminator with
    | Ret opn ->
      Asm.[Movq, [codegen_operand opn; ~%Rax]
          ;Jmp, [~$$"exit"]]
    | Br lbl -> Asm.[Jmp, [~$$lbl]]
    | Cbr (cond, opn, then_lbl, else_lbl) ->
      let cc = match cond with
        | EqZ -> X86.Eq
        | LeZ -> X86.Le
        | LtZ -> X86.Lt
      in
      Asm.[Cmpq, [~$0; codegen_operand opn]
          ;J cc, [~$$then_lbl]
          ;Jmp, [~$$else_lbl]]
  in
  (List.flatten (List.map (codegen_instr c) block.instrs))@terminator

let codegen_cfg (ctx : IR.ctx) (cfg : IR.cfg) =
  let open IR in
  (* If label is unvisited, compute Some trace beginning with label;
     otherwise return None *)
  let rec continue_trace label blocks =
    match List.assoc_opt label blocks with
    | Some block ->
      Some (codegen_trace label block (List.remove_assoc label blocks))
    | None -> None

  (* Generate code for a trace beginning with (label, block).  Return
     a pair consisting of the code for the trace and a list of
     *unvisited* blocks *)
  and codegen_trace label block blocks =
    let asm_instrs =
      List.flatten (List.map (codegen_instr ctx) block.instrs)
    in
    let open Asm in
    match block.terminator with
    | Ret opn ->
      let asm_block =
        Asm.text label @@
        asm_instrs
        @[Movq, [codegen_operand opn; ~%Rax]
         ;Jmp, [~$$"exit"]]
      in
      ([asm_block], blocks)
    | Br target ->
      begin match continue_trace target blocks with
        | Some (trace, blocks) ->
          let asm_block =
            Asm.text label asm_instrs
          in
          (asm_block::trace, blocks)
        | None ->
          let asm_block =
            Asm.text label (asm_instrs@[Jmp, [~$$ target]])
          in
          ([asm_block], blocks)
      end
    | Cbr (cond, opn, then_lbl, else_lbl) ->
      let cc = match cond with
        | EqZ -> X86.Eq
        | LeZ -> X86.Le
        | LtZ -> X86.Lt
      in
      begin match continue_trace else_lbl blocks with
        | Some (trace, blocks) ->
          let asm_block =
            Asm.text label @@
            asm_instrs
            @[Cmpq, [~$0; codegen_operand opn]
             ;J cc, [~$$then_lbl]]
          in
          (asm_block::trace, blocks)
        | None ->
          match continue_trace then_lbl blocks with
          | Some (trace, blocks) ->
            let neg_cc = match cc with
              | Eq -> Neq
              | Neq -> Eq
              | Gt -> Le
              | Ge -> Lt
              | Lt -> Ge
              | Le -> Gt
            in
            let asm_block =
              Asm.text label @@
              asm_instrs
              @[Cmpq, [~$0; codegen_operand opn]
               ;J neg_cc, [~$$else_lbl]]
            in
            (asm_block::trace, blocks)
          | None ->
            let asm_block =
              Asm.text label @@
              asm_instrs
              @[Cmpq, [~$0; codegen_operand opn]
               ;J cc, [~$$then_lbl]
               ;Jmp, [~$$else_lbl]]
            in
            ([asm_block], blocks)
      end
  in
  (* Compute a covering set of traces, beginning with a trace that
     starts at (label, block) *)
  let rec codegen_blocks label block blocks =
    let (trace, blocks) = codegen_trace label block blocks in
    match blocks with
    | [] -> trace (* Done -- all blocks are covered *)
    | (lbl,blk)::blocks ->
      let rest = (codegen_blocks lbl blk blocks) in
      trace @ rest
  in
  let rec push_zeros n =
    if n == 0 then []
    else Asm.((Pushq, [~$0])::(push_zeros (n-1)))
  in
  let prologue =
    Asm.[Pushq, [~%Rbp]           (* save frame pointer *)
        ;Movq, [~%Rsp; ~%Rbp]     (* create new frame pointer *)
        (* allocate local storage *)
        ;Subq, [~$(( (IR.nb_uid ctx)) * 8); ~%Rsp] 
        ]@(push_zeros (IR.nb_vars ctx))
  in

  let epilogue =
    Asm.(text "exit"
           (Asm.[(* deallocate local storage *)
                 Addq, [~$(((IR.nb_uid ctx) + (IR.nb_vars ctx)) * 8); ~%Rsp]
                ;Popq, [~%Rbp] (* restore frame pointer *)
                ;Retq, []]))   (* return control to caller *)
  in
  (Asm.gtext "program" prologue
   :: codegen_blocks "dummy" (fst cfg) (snd cfg)
   @[epilogue])


let codegen_cfg (c : IR.ctx) (cfg : IR.cfg) =
  let rec push_zeros n =
    if n == 0 then []
    else Asm.((Pushq, [~$0])::(push_zeros (n-1)))
  in
  let prologue =
    Asm.[Pushq, [~%Rbp]           (* save frame pointer *)
        ;Movq, [~%Rsp; ~%Rbp]     (* create new frame pointer *)
        (* allocate local storage *)
        ;Subq, [~$(( (IR.nb_uid c)) * 8); ~%Rsp] 
        ]@(push_zeros (IR.nb_vars c))
  in
  let labelled_blocks =
    List.map (fun (label, block) -> Asm.text label (codegen_block c block)) (snd cfg)
  in
  let epilogue =
    Asm.(text "exit"
           (Asm.[(* deallocate local storage *)
                 Addq, [~$(((IR.nb_uid c) + (IR.nb_vars c)) * 8); ~%Rsp]
                ;Popq, [~%Rbp] (* restore frame pointer *)
                ;Retq, []]))   (* return control to caller *)
  in
  ((Asm.gtext "program" (prologue @ (codegen_block c (fst cfg))))::labelled_blocks)
  @[epilogue]

let compile (ast : Ast.cmd) : X86.prog =
  let vars = Ast.var_list ast in
  let c = IR.mk_context vars in
  let stream =
    translate_cmd c ast
    >:: IR.(T (Ret (Const 0L)))
  in
  let cfg = build_cfg stream in
  codegen_cfg c cfg
